from __future__ import annotations

import os
import json

from typing import Dict, Any

from .model import (
    TrelloCards
)
from .controller import (
    ingest_board_data,
)


def get_trello_board_dict_from_json(trello_file_path: str) -> dict:
    trello_data: Dict[str, Any] = {}
    trello_json_path = os.path.join(
        os.path.dirname(__file__),
        "json_data",
        trello_file_path
    )
    with open(trello_json_path, "r") as fin:
        trello_data = json.loads(fin.read())
    return trello_data


def get_cards_in_list(trello_data: dict, list_name: str) -> TrelloCards:
    trello_instance = ingest_board_data(trello_data)
    return [card for card in trello_instance.cards if card.card_list.name == list_name]


"""
Likely Ingestion Process:

1. JSON -> TrelloBoard
2. TrelloBoard.cards -> .md + .md5 checksum of desc
3. .md -> .html

During Re-ingestion:
- If .md file exists
- generate and compare .md5 files
- If .md5 files differ, generate new .md + .md5 file
- Else do nothing

Possible Enhancement:

content
YYYY-MM-DD/
    file1.md
    file2.md
    file1.html.checksum
    file2.html.checksum

Generate HTML file in memory
Generate Checksum in memory
Compare checksum against existing one,
    if no existing checksum or they differ
        generate HTML file
        generate checksum file
"""



import os
import re
import yaml
import http.server
import socketserver
from typing import Generator, List
import markdown
from collections import namedtuple
import logging
from jinja2 import Environment, FileSystemLoader

from .utils import mkdirp
from .meta_files.model import (
    FileContext,
    FileContextBody,
    FileContextMeta,
    Md,
)

template_path = os.path.join(
    os.path.dirname(__file__),
    'theme',
    'templates',
)
template_loader = FileSystemLoader([template_path])

env = Environment(loader=template_loader)
# autoescape=select_autoescape(['html'])


logger = logging.getLogger('basicLogger')
logger.setLevel(logging.DEBUG)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(
    logging.Formatter('%(message)s')
)
logger.addHandler(console_handler)

default_markdown_parser = markdown.Markdown(
    extensions=[
        "markdown.extensions.tables",
        "markdown.extensions.admonition",
        "markdown.extensions.smarty",
        "pymdownx.magiclink",
        "pymdownx.betterem",
        "pymdownx.tilde",
        "pymdownx.emoji",
        "pymdownx.tasklist",
        "pymdownx.superfences",
        "pymdownx.highlight",
        "pyembed.markdown",  # https://pyembed.github.io/usage/markdown/
    ],
    extension_configs={},
)


def traverse_files_in_path(folder_path) -> Generator[str, None, None]:
    for root, dirs, files in os.walk(folder_path):
        if not files:
            continue
        for file in files:
            yield os.path.join(root, file)


def split_meta_from_markdown(markdown_text) -> Md:
    if markdown_text.startswith("---"):
        try:
            return Md(*markdown_text.split("---")[1:3])
        except IndexError:
            pass
    return Md("", markdown_text)


def extract_meta_from_markdown_text(
    markdown_text: str, meta_parser=yaml.safe_load
) -> dict:
    return meta_parser(split_meta_from_markdown(markdown_text).yaml_meta)


def generate_html_from_markdown_content(
    markdown_text: str, markdown_parser=default_markdown_parser
) -> str:
    return markdown_parser.convert(
        split_meta_from_markdown(markdown_text).markdown_content
    )


def parse_markdown_file(file_path, encoding="utf-8") -> FileContext:
    with open(file_path, "rb") as fin:
        markdown_text = fin.read().decode(encoding)

        context_meta = FileContextMeta(extract_meta_from_markdown_text(markdown_text))
        context_body = FileContextBody(
            generate_html_from_markdown_content(markdown_text)
        )
        return FileContext(meta=context_meta, content=context_body)


def file_context_to_html(context: FileContext) -> str:
    template = env.get_template(context.get_template())
    return template.render(context.get_context())


def generate_html_from_markdown_file(
    file_path: str, source_dir: str, dest_dir: str
) -> None:
    file_context = parse_markdown_file(file_path)
    if not file_context.content.content:
        return

    html_content = file_context_to_html(file_context)
    relative_folder_path = os.path.dirname(file_path)[len(source_dir):]
    if relative_folder_path:
        if relative_folder_path.startswith('/'):
            relative_folder_path = relative_folder_path[1:]
        if relative_folder_path[-1] != '/':
            relative_folder_path = relative_folder_path + '/'

    output_file_name = re.sub(r"\.md$", ".html", os.path.basename(file_path))
    output_file_folder = os.path.join(dest_dir, relative_folder_path)
    mkdirp(output_file_folder)
    build_path = os.path.join(
        output_file_folder,
        output_file_name
    )
    with open(build_path, "wb") as fout:
        fout.write(html_content.encode("utf-8"))


def build(src_dir: str, dest_dir: str) -> None:
    for file_path in traverse_files_in_path(src_dir):
        generate_html_from_markdown_file(file_path, src_dir, dest_dir)


def serve(dest_dir: str) -> None:
    LOCALHOST = "127.0.0.1"
    PORT = int(os.environ.get("BLOG_PORT", 8000))
    web_dir = os.path.join(os.path.dirname(__file__), dest_dir)
    os.chdir(web_dir)
    Handler = http.server.SimpleHTTPRequestHandler
    httpd = socketserver.TCPServer(("127.0.0.1", PORT), Handler)
    print(f"serving http://{LOCALHOST}:{PORT}")
    httpd.serve_forever()


FolderMeta = namedtuple('FolderMeta', ['dir_path', 'files', 'has_dirs'])


def _get_tail_nodes_in_folder(dest_dir) -> List[FolderMeta]:
    """
    Traverse through a folder and return only FolderMeta
    of files without child folders
    """
    folders = []
    for root, dirs, files in os.walk(dest_dir):
        folders.append(
            FolderMeta(
                dir_path=root,
                files=[os.path.join(root, file) for file in files],
                has_dirs=bool(len(dirs))))
    return [i for i in folders if not i.has_dirs]


def clean(dest_dir: str) -> None:
    """
    1. Traverse through folder finding folders without child folders
    2. Delete "tail" folders and their files
    3. Repeat until all "child" folders of "dest_dir" are deleted

    If _get_tail_nodes_in_folder is called 100 times print error message
    """
    for i in range(100):
        folders_to_delete = _get_tail_nodes_in_folder(dest_dir)
        if not folders_to_delete:
            return
        for folder in folders_to_delete:
            if folder.dir_path == dest_dir:
                return
            for file in folder.files:
                if ".keep" in file:
                    continue
                os.remove(file)
            if ".keep" not in [os.path.basename(f) for f in folder.files]:
                os.removedirs(folder.dir_path)
    else:
        raise Exception("Too many sub-folder levels")



import os
import re
import codecs


def mkdirp(dir_path) -> None:
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def slugify(string) -> str:
    _string = string
    unicode_quote_chars = [
        '„',
        '‚',
        '“',
        '‟',
        '‘',
        '‛',
        '”',
        '’',
        '"',
        '❛',
        '❜',
        '❟',
        '❝',
        '❞',
        '❮',
        '❯',
        '⹂',
        '〝',
        '〞',
        '〟',
        '＂',
        '«',
        '‹',
        '»',
        '›',
    ]
    for char in unicode_quote_chars:
        _string = re.sub(char, '', _string)
    _string = re.sub(r',|\:|\?|\(|\)|\'|\!|<|>', '', _string)

    _string = re.sub(r'&', 'and', _string)
    _string = re.sub(r'%', 'percent', _string)

    # Unicode small space is not visible, and odd with monotype editor fonts
    # so its encoded in bytes for the sake of making sense of this following line
    unicode_small_space = codecs.decode(b'\xe2\x80\x8a', 'utf-8')
    unicode_dash_chars = [
        '—',
        '‒',
        '–',
        '⁓',
        '┄',
        '﹉',
        '╍',
        '﹍',
        '┅',
        '┈',
        '┉',
        '┉',
        '┉',
        unicode_small_space,
    ]
    _string = re.sub(r'\ |\/', '-', _string)
    for char in unicode_dash_chars:
        _string = re.sub(char, '-', _string)

    if _string.endswith('.'):
        _string = _string[:-1]
    return _string.lower()



from __future__ import annotations

from datetime import datetime
import dataclasses
import pprint
from typing import List, Optional, Generator


@dataclasses.dataclass
class TrelloLabel:
    id: str
    idBoard: str
    name: str
    color: str


TrelloLabels = List[TrelloLabel]


@dataclasses.dataclass
class TrelloList:
    id: str
    name: str
    closed: bool
    idBoard: str
    pos: int
    subscribed: bool


TrelloLists = List[TrelloList]


@dataclasses.dataclass
class TrelloCard:
    id: str
    dateLastActivity: datetime
    desc: str
    idBoard: str
    idLabels: str
    idList: str
    name: str
    pos: int
    labels: Optional[TrelloLabels]
    card_list: TrelloList


TrelloCards = List[TrelloCard]


class TrelloBoard:
    def __init__(
        self,
        labels: TrelloLabels = None,
        lists: TrelloLists = None,
        cards: TrelloCards = None,
    ):
        self._labels = labels or []
        self._lists = lists or []
        self._cards = cards or []

    @property
    def labels(self) -> TrelloLabels:
        return self._labels

    @property
    def lists(self) -> TrelloLists:
        return self._lists

    @property
    def cards(self) -> TrelloCards:
        return self._cards

    @staticmethod
    def _gen_filter_labels(label_names: List[str], labels: TrelloLabels) -> Generator[TrelloLabel, None, None]:
        for label_obj in labels:
            if label_obj.name in label_names:
                yield label_obj

    @staticmethod
    def filter_labels(label_names: List[str], labels: TrelloLabels) -> Optional[TrelloLabels]:
        return [label_obj for label_obj in TrelloBoard._gen_filter_labels(label_names, labels)]

    def pretty_print(self) -> None:
        print("labels")
        pprint.pprint(self.labels)
        print("\nlists")
        pprint.pprint(self.lists)
        print("\ncards")
        pprint.pprint(self.cards)



#!/usr/bin/env python

"""blog

Usage:
  blog build <src> <dest>
  blog clean <dest>
  blog serve <dest>
  blog trello <json_file> <list_name> [--write]
  blog (-h | --help)
  blog (-v | --version)

Options:
  -h --help     Show this screen.
  -v --version  Show version.
  --write       Write to disk
"""

import os
import sys
from docopt import docopt

# Not Sure why mypy complains about missing attributes on __main__.blog
# for now ignoring mypy lint errors
from ._blog import (  # type: ignore
    serve,
    build,
    clean,
)
from .trello_to_markdown import (
    trello_json_export_to_markdown_files,
)

COMMAND_BUILD = "build"
COMMAND_CLEAN = "clean"
COMMAND_SERVE = "serve"
COMMAND_TRELLO = "trello"
SOURCE = "<src>"
DESTINATION = "<dest>"


def get_command(arguments: dict) -> str:
    commands = [COMMAND_SERVE, COMMAND_BUILD, COMMAND_CLEAN, COMMAND_TRELLO]
    return next((k for k, v in arguments.items() if v and k in commands))


VERSION = ''
version_file = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'version'
)
with open(version_file, 'r', encoding='utf-8') as fin:
    VERSION = fin.read().strip()


if __name__ == "__main__":
    arguments = docopt(__doc__, version=f'blog {VERSION}')
    command = get_command(arguments)
    if command == COMMAND_BUILD:
        src_dir = os.path.abspath(arguments[SOURCE])
        dest_dir = os.path.abspath(arguments[DESTINATION])
        build(src_dir=src_dir, dest_dir=dest_dir)
    elif command == COMMAND_CLEAN:
        dest_dir = os.path.abspath(arguments[DESTINATION])
        try:
            clean(dest_dir=dest_dir)
        except Exception as exp:
            sys.stderr.write(str(exp))
    elif command == COMMAND_SERVE:
        dest_dir = os.path.abspath(arguments[DESTINATION])
        serve(dest_dir=dest_dir)
    elif command == COMMAND_TRELLO:
        trello_file = os.path.abspath(arguments['json_file'])
        trello_json_export_to_markdown_files(
            trello_file=trello_file,
            trello_list_name=arguments['list_name'],
            write_to_disk=arguments['write'])
    else:
        print(__doc__)



from __future__ import annotations

from datetime import datetime
from typing import Any

from .model import (
    TrelloBoard, TrelloLabel, TrelloLabels, TrelloList, TrelloLists, TrelloCard, TrelloCards
)


def _force_string(thing: Any) -> str:
    if isinstance(thing, bytes):
        return thing.encode('utf-8')  # type: ignore
    return thing


def ingest_label(label_data: dict) -> TrelloLabel:
    return TrelloLabel(**label_data)


def ingest_labels(board_data: dict) -> TrelloLabels:
    return [ingest_label(label_data) for label_data in board_data["labels"]]


def ingest_list(list_data: dict) -> TrelloList:
    return TrelloList(
        id=list_data['id'],
        name=list_data['name'],
        closed=list_data['closed'],
        idBoard=list_data['idBoard'],
        pos=list_data['pos'],
        subscribed=list_data['subscribed'])


def ingest_lists(board_data: dict) -> TrelloLists:
    return [ingest_list(list_data) for list_data in board_data["lists"]]


def ingest_card(card_data: dict, lists: TrelloLists, labels: TrelloLabels = None) -> TrelloCard:
    card_list = next((l for l in lists if l.id == card_data['idList']))
    return TrelloCard(
        id=card_data['id'],
        dateLastActivity=datetime.strptime(
            card_data["dateLastActivity"],
            "%Y-%m-%dT%H:%M:%S.%fZ",
        ),
        desc=card_data['desc'],
        idBoard=card_data['idBoard'],
        idLabels=card_data['idLabels'],
        idList=card_data['idList'],
        name=card_data['name'],
        pos=card_data['pos'],
        labels=labels,
        card_list=card_list,
    )


def ingest_cards(
    board_data: dict, lists: TrelloLists, labels: TrelloLabels = None
) -> TrelloCards:
    return ([ingest_card(card_data, lists, labels) for card_data in board_data["cards"]])


def ingest_board_data(board_data: dict) -> TrelloBoard:
    labels = ingest_labels(board_data)
    lists = ingest_lists(board_data)
    cards = ingest_cards(board_data, lists, labels)
    return TrelloBoard(
        labels=labels,
        lists=lists,
        cards=cards,
    )



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

VERSION = ''
version_file = os.path.join(
    here,
    'blog',
    'version'
)
with open(version_file, 'r', encoding='utf-8') as fin:
    VERSION = fin.read().strip()

setup(
    name='blog',
    version=VERSION,
    description='A static website generator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tomleo/blog',
    author='Tom Leo',
    author_email='tom@tomleo.com',
    # See https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
    keywords='static website generator',
    packages=find_packages(exclude=['data', 'tests']),
    python_requires='>=3.7.*',
    # Top level requires are defined in requirements.in
    # Resolved minimum versions are the result of pip-compile with py3.7
    # It's possible older versions of these libraries would be compatible
    setup_requires=[
        'wheel'
    ],
    install_requires=[
        'PyYAML>=5.1.*',
        'Jinja2>=2.10.*',
        'Markdown>=3.1.*',
        'Pygments>=2.4.*',
        'pymdown-extensions>=6.*',
        'pyembed-markdown>=1.1.*',
        'docopt>=0.6.*',
    ],
    extras_require={
        'dev': [
            'black',
            'flake8',
            'mypy',
            'isort'
        ],
        'test': [
            'pytest',
        ]
    },
    entry_points={
        'console_scripts': [
            'blog=blog:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/tomleo/blog/issues',
        'Funding': 'https://donate.pypi.org',
        'Source': 'https://github.com/tomleo/blog/',
    },
)



import os
import yaml
from datetime import datetime

from .trello.view import get_trello_board_dict_from_json, get_cards_in_list
from .trello.model import TrelloCard
from .utils import mkdirp, slugify


def create_markdown_file_from_trello_card(
    card: TrelloCard, trello_folder_name: str = "content/trello"
) -> None:
    labels = card.labels or []
    if labels:
        card_folder_name = slugify(labels[0].name)
    else:
        card_folder_name = "unclassified"
    card_folder_path = os.path.join(
        os.path.dirname(__file__),
        trello_folder_name,
        card_folder_name,
        datetime.strftime(card.dateLastActivity, "%Y-%m-%d"),
    )
    mkdirp(card_folder_path)
    # TODO: add lastmodified, labels, ect.
    markdown_meta = yaml.dump({"title": card.name})
    markdown_content = "---\n{}\n---\n{}".format(markdown_meta, card.desc)
    markdown_file_name = "{}.md".format(slugify(card.name))
    markdown_file_path = os.path.join(card_folder_path, markdown_file_name)
    with open(markdown_file_path, "wb") as fout:
        fout.write(markdown_content.encode("utf-8"))


def trello_json_export_to_markdown_files(
    trello_file: str, trello_list_name: str, write_to_disk: bool = True
):
    """
    Reads trello board JSON export, and converts card into markdown files
    """

    trello_file_path = os.path.abspath(trello_file)
    trello_data = get_trello_board_dict_from_json(trello_file_path)
    read_cards = get_cards_in_list(trello_data, trello_list_name)
    if write_to_disk:
        for card in read_cards:
            create_markdown_file_from_trello_card(card)



import dataclasses
from typing import NamedTuple


class FileContextMeta:
    TEMPLATE_STR = "template"
    DEFAULT_TEMPLATE_STR = "default"

    def __init__(self, meta: dict = None):
        if not meta:
            meta = {}
        self._template = meta.get(self.TEMPLATE_STR, self.DEFAULT_TEMPLATE_STR)
        self._meta = {k: v for k, v in meta.items() if k != self.TEMPLATE_STR}

    @property
    def template(self) -> str:
        return self._template

    def items(self):
        yield (self.TEMPLATE_STR, self.template)
        for k, v in self._meta.items():
            yield (k, v)

    def __getitem__(self, key):
        if key == self.TEMPLATE_STR:
            return self.template
        return self._meta[key]

    def __setitem__(self, key, val):
        if key == self.TEMPLATE_STR:
            self._template = val
        else:
            self._meta[key] = val

    def as_dict(self) -> dict:
        return {k: v for (k, v) in self.items()}


@dataclasses.dataclass
class FileContextBody:
    content: str


@dataclasses.dataclass
class FileContext:
    meta: FileContextMeta
    content: FileContextBody

    def get_template(self) -> str:
        return f"{self.meta.template}.html"

    def get_context(self) -> dict:
        return {
            "meta": self.meta.as_dict(),
            "content": self.content.content
        }


class Md(NamedTuple):
    yaml_meta: str
    markdown_content: str
